/* 
 * The MIT License (MIT)
 *
 * Copyright (c) 2019 Ha Thach (tinyusb.org)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * This version converts the demo program to use protothreads 1.3
 * then factors out the reporting from the usb handler and controls
 * the mouse and keyboard disply to VGA
 * The USB task is running as a protothread every 8 mSec to catch mouse events
 * YOu might need ot speed this up for gaming.
 * core 0 is handling USB and blinking an LED
 * core 1 is doing the drawing
 * 
 * // modify ff.c line 6848
   // turn on f_gets option in ffconf.h
   // modify hid_app (my_hid_app) to copy out mouse and keyboard events
   // modify msc_app to disable CLI command interface (I like mine better)
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "vars.h"
#include "bsp/board.h"
#include "tusb.h"

#include "C:\Users\16073\OneDrive\Documents\Pico-v1.5.0\pico-sdk\lib\tinyusb\lib\fatfs\source\ff.h"
#include "C:\Users\16073\OneDrive\Documents\Pico-v1.5.0\pico-sdk\lib\tinyusb\lib\fatfs\source\diskio.h"

// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_3.h"
// global vars shared with HID code thru vars.h
int mouse_dx, mouse_x=320, mouse_dy, mouse_y=240 ;
int mouse_update ;
int mouse_button, mouse_button_update ;
char keybd_char ;
int keybd_update ;

// user modifiable USB tasks
extern void cdc_app_task(void);
extern void hid_app_task(void);
extern void msc_app_task(void);
extern void msc_app_init(void);

//--------------------------------------------------------------------+
// TinyUSB Callbacks
//--------------------------------------------------------------------+

void tuh_mount_cb(uint8_t dev_addr)
{
  // application set-up
  printf("A device with address %d is mounted\r\n", dev_addr);
}

void tuh_umount_cb(uint8_t dev_addr)
{
  // application tear-down
  printf("A device with address %d is unmounted \r\n", dev_addr);
}

// ==================================================
// === USB hid thread on core 0
// ==================================================
// handle mouaswe and keybd
static PT_THREAD (protothread_hid(struct pt *pt))
{
    PT_BEGIN(pt);
     while (1) {
      // tinyusb host task
      tuh_task();
      //cdc_app_task();
        //
      // uncomment for status commands
      //hid_app_task(); 
      //
      // uncomment for disk menu system
      //msc_app_task();
      //
      // 8 mSec 125 Hz
      //sleep_ms(1);
      PT_YIELD_usec(8000) ;
    }
    PT_END(pt);
}

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.5 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(500000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ===========================================
// serial and file i/o
// ===========================================
// The command interpreter is largely copied from
// the tinyusb disribution.
// See license at top of file
// ===========================================
static PT_THREAD (protothread_file(struct pt *pt))
{
    PT_BEGIN(pt);
    static char serial_buffer[40];
    static char cmd[16], arg1[64], arg2[16], arg3[16] ;
    static char* token ;
    float farg ;
    int arg ;
    //
    while(1) {
        printf(">>");
         // spawn a thread to do the non-blocking serial read
         serial_read ;
        // tokenize
        token = strtok(pt_serial_in_buffer, "  ");
        strcpy(cmd, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg1, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg2, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg3, token) ;
        // ===
        if(strcmp(cmd,"help")==0){
            printf("*****\n\r");
            printf("write filename range step -- demo data write -- ints to range\n\r") ;
            printf("plot filename -- data to VGA") ;
            printf("cd directory -- changes current dir\n\r");
            printf("rm filename -- delinks a file\n\r");
            printf("cat filename -- prints file\n\r");
            printf("ls directory -- list dir contents\n\r");
            printf("mkdir directory -- new dir\n\r");
            printf("*****\n\r");
        }

        // ===
        if(strcmp(cmd,"write")==0){
            FIL f_dst;
            if ( FR_OK != f_open(&f_dst, arg1, FA_WRITE | FA_CREATE_ALWAYS) )
            {
                printf("cannot create '%s'\r\n", arg1);
            }else {
                uint8_t buf[32] = "Data from Pico\n\r";
                UINT wr_count = 0;
                if ( FR_OK != f_write(&f_dst, buf, strlen(buf), &wr_count) )
                {
                    printf("cannot write to '%s'\r\n", arg1);
                }
                uint32_t range, step;
                sscanf(arg2, "%d", &range) ;
                sscanf(arg3, "%d", &step) ;
                for(int i=0; i<=range; i+=step){
                    uint8_t data[10];
                    sprintf(data,"%d\n", i);
                    f_write(&f_dst, data, strlen(data), &wr_count);
                }
            }
             f_close(&f_dst);
        }
        
        // ===
        if(strcmp(cmd,"cd")==0){
            // default is current directory
            const char* dpath = arg1;

            if ( FR_OK != f_chdir(dpath) )
            {
                printf("%s: No such file or directory\r\n", dpath);
            }
        }

        // ===
        if(strcmp(cmd,"plot")==0){
            FIL fi;
            if ( FR_OK != f_open(&fi, arg1, FA_READ) ) {
                printf("%s: No such file or directory\r\n", arg1);
            }
            else {
                uint8_t buf[100];
                UINT count = 0;
                // modify ff.c line 6848
                // turn on f_gets option in ffconf.h
                //
                // get number of data points
                int file_count ;
                f_gets(buf, sizeof(buf), &fi ) ;
                sscanf(buf, "%d", &file_count);
                // read and plot
                for(int i=0; i<=file_count; i++) 
                {
                    float vga_x, vga_y ;
                    f_gets(buf, sizeof(buf), &fi ) ;
                    //printf("%s", buf);
                    sscanf(buf, "%f %f", &vga_x, &vga_y);
                    drawPixel((short)vga_x, (short)vga_y, CYAN) ;
                }
            }
            f_close(&fi);
        }


        // ===
        if(strcmp(cmd,"cat")==0){
            FIL fi;
            if ( FR_OK != f_open(&fi, arg1, FA_READ) ) {
                printf("%s: No such file or directory\r\n", arg1);
            }
            else {
                uint8_t buf[512];
                UINT count = 0;
                while ( (FR_OK == f_read(&fi, buf, sizeof(buf), &count)) && (count > 0) )
                {
                    for(UINT c = 0; c < count; c++)
                    {
                        const uint8_t ch = buf[c];
                        putchar(ch);
                    }
                }
                printf("\n\r") ;
            }
            f_close(&fi);
        }

        // ===
        if(strcmp(cmd,"ls")==0){
            // default is current directory
            const char* dpath = ".";
            if (arg1) dpath = arg1;

            DIR dir;
            if ( FR_OK != f_opendir(&dir, dpath) )
            {
                printf("cannot access '%s': No such file or directory\r\n", dpath);
            }

            FILINFO fno;
            while( (f_readdir(&dir, &fno) == FR_OK) && (fno.fname[0] != 0) )
            {
                if ( fno.fname[0] != '.' ) // ignore . and .. entry
                {
                if ( fno.fattrib & AM_DIR )
                {
                    // directory
                    printf("/%s\r\n", fno.fname);
                }else
                {
                    printf("%-40s", fno.fname);
                    if (fno.fsize < 1024)
                    {
                    printf("%lu B\r\n", fno.fsize);
                    }else
                    {
                    printf("%lu KB\r\n", fno.fsize / 1024);
                    }
                }
                }
            }
            f_closedir(&dir);
        }
        
        // ===
        if(strcmp(cmd,"rm")==0){
            const char* fpath = arg1 ; // token count from 1
            if ( FR_OK != f_unlink(fpath) )
            {
            printf("cannot remove '%s': No such file or directory\r\n", fpath);
            }
        }

        // ===
        if(strcmp(cmd,"mkdir")==0){
            const char* dpath = arg1;
            if ( FR_OK != f_mkdir(dpath) )
            {
                printf("%s: cannot create this directory\r\n", dpath);
            }
        }
        
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ===========================================
// Cursor  -- overlay conttol
// ===========================================

short cursor_pixel_x[5], cursor_pixel_y[5] ;


// replace cursor color with saved image colors
void erase_cursor(short x, short y){
    // first the x direction, then y
    for(int i=0; i<5; i++){
        drawPixel(x-2+i, y,  cursor_pixel_x[i]) ;
        drawPixel(x, y-2+i,  cursor_pixel_y[i]) ;
    }
}

// replace image with cursor color and save image pixels
void draw_cursor(short x, short y){
    for(int i=0; i<5; i++){
        cursor_pixel_x[i] = readPixel(x-2+i, y) ;
        cursor_pixel_y[i] = readPixel(x, y-2+i) ;
        drawPixel(x-2+i, y, WHITE) ;
        drawPixel(x, y-2+i, WHITE) ;
    }
}

// ==================================================
// === vga thread on core 1
// ==================================================
//vga display of mouse and keybd 
static PT_THREAD (protothread_vga(struct pt *pt))
{
    PT_BEGIN(pt);
    static char video_str[64], mouse_str[10];
    static unsigned char draw_color = GREEN ;
    static int text_x, text_y, begin_text_x ;

    // Write some text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeStringBold("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeStringBold("USB HID") ;
    setCursor(65, 20) ;
    writeString("Hunter Adams") ;
    setCursor(65, 30) ;
    writeString("Bruce Land") ;
    setCursor(65, 40) ;
    writeStringBold("ece4760 Cornell") ;
    //
    setCursor(445, 10) ;
    setTextColor(WHITE) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.3") ;
    setCursor(445, 20) ;
    writeString("VGA 4-bit color") ;
    setCursor(445, 30) ;
    writeString("Tested on pico") ;
    //
    setCursor(270, 20) ;
    setTextColor2(WHITE, BLACK) ;
    writeStringBig("USB HID Test") ;
    setCursor(270, 37) ;
    setTextColor2(WHITE, BLACK) ;
    
    draw_cursor(mouse_x, mouse_y);
    drawRect(2,90,636, 389, WHITE) ;
    setCursor(290, 92) ;
    setTextColor2(WHITE, BLACK) ;
    sprintf(video_str, "%s","Draw Area") ;
    writeStringBold(video_str) ;
    drawHLine(2,100,636, WHITE) ;

     while (1) {
      
      // 30 mSec
      PT_YIELD_usec(30000) ;
     // dreaw the readouts and controls
     setCursor(300, 73) ;
    setTextColor2(GREEN, BLACK) ;
    sprintf(video_str, "%s","Color pick (middle click)") ;
     writeStringBold(video_str) ;
     for(int i=0; i<16; i++){
      fillRect(300+10*i,60,10,10, i);
     }

     drawRoundRect(500, 60, 50, 15, 7, WHITE);
     setCursor(506, 64) ;
     setTextColor2(RED, BLACK) ;
     writeStringBold("erase") ;
     if(mouse_x>500 && mouse_x<610 & mouse_y>60 & mouse_y<75 && (mouse_button & 1)) fillRect(3,101,634, 377, BLACK);

      setCursor(20, 60) ;
      setTextColor2(GREEN, BLACK) ;
      sprintf(video_str, "mouse x = %03d   mouse y = %03d", mouse_x, mouse_y) ;
     writeStringBig(video_str) ;

     // cursor motion update
     if(mouse_update){
      erase_cursor(mouse_x, mouse_y);
      mouse_x += mouse_dx ;
      if (mouse_x < 5) mouse_x = 5;
      if (mouse_x > 635) mouse_x = 635;
      mouse_y += mouse_dy ;
      if (mouse_y < 5) mouse_y = 5;
      if (mouse_y > 475) mouse_y = 475;
      mouse_update = 0 ;
      draw_cursor(mouse_x, mouse_y);
     }

     // mouse buttons
     // left -- draw
     // right -- text instetion point
     // middle -- grab color fro mouse position
     if(mouse_button_update || mouse_button==0){
      //
        if(mouse_button & 1) {
          strcpy(mouse_str, "left    ");
          //2,90,636, 389,
          if(mouse_x>2 && mouse_x<638 && mouse_y>101 && mouse_y<478){
            fillCircle(mouse_x, mouse_y, 1, draw_color) ;
            for(int i=1; i<4; i++){
              cursor_pixel_x[i] = draw_color ;
              cursor_pixel_y[i] = draw_color ;
            }  
          }      
        }    
        // 
        else if(mouse_button & 2) {
          strcpy(mouse_str, "right   ");
          text_x = mouse_x;
          text_y = mouse_y;
          begin_text_x = mouse_x ;
        }
        //
        else if(mouse_button & 4) {
          strcpy(mouse_str, "middle  ");
          draw_color = readPixel(mouse_x+1, mouse_y+1) ;
        }
        //
        else strcpy(mouse_str, "none    ");
        setCursor(20, 75) ;
        setTextColor2(GREEN, BLACK) ;
        sprintf(video_str, "mouse button = %s  ", mouse_str) ;
        writeStringBig(video_str) ;
        mouse_button_update = false ;
     } 

    // keyboard event
    if (keybd_update){
      // <enter>
      if(keybd_char == '\r') {
        text_y += 15; 
        text_x = begin_text_x; 
      }
      // delete a char on the same line
      // <backspace>
      else if(keybd_char == '\b') {
        text_x -= 8 ;
        // draw a spacw over the previous char
        drawCharBig(text_x, text_y, 0x20, WHITE, BLACK) ;
      }
      // regular char
      else {
        drawCharBig(text_x, text_y, keybd_char, draw_color, BLACK) ;
        text_x += 8 ; 
      }   
      keybd_update = false ;
    }    
    }
    PT_END(pt);
}

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_vga) ;
  pt_add_thread(protothread_file) ;
  //
  // === initalize the scheduler ==========
  pt_sched_method = SCHED_ROUND_ROBIN ;

  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){

  board_init();

  printf("TinyUSB Host HID -- Protothreads 1.3\r\n");

  // init host stack on configured roothub port
  tuh_init(BOARD_TUH_RHPORT);
  // init needs some time
//    if (board_init_after_tusb) {
//     board_init_after_tusb();
//   }
  sleep_ms(500);

   msc_app_init();
   sleep_ms(200);
  

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_hid);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_sched_method = SCHED_PRIORITY ;
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main